/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.builders.snapshot;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.annotation.Nonnull;

import org.apache.commons.lang3.tuple.Pair;

import net.minecraft.block.Block;
import net.minecraft.block.BlockFalling;
import net.minecraft.block.properties.IProperty;
import net.minecraft.block.state.IBlockState;
import net.minecraft.init.Blocks;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTBase;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTUtil;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.ResourceLocation;
import net.minecraft.util.Rotation;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;

import net.minecraftforge.fluids.FluidStack;

import buildcraft.api.core.InvalidInputDataException;
import buildcraft.api.schematics.ISchematicBlock;
import buildcraft.api.schematics.SchematicBlockContext;

import buildcraft.lib.misc.BlockUtil;
import buildcraft.lib.misc.NBTUtilBC;

public class SchematicBlockDefault implements ISchematicBlock {
    @SuppressWarnings("WeakerAccess")
    protected final Set<BlockPos> requiredBlockOffsets = new HashSet<>();
    @SuppressWarnings("WeakerAccess")
    protected IBlockState blockState;
    @SuppressWarnings("WeakerAccess")
    protected final List<IProperty<?>> ignoredProperties = new ArrayList<>();
    @SuppressWarnings("WeakerAccess")
    protected NBTTagCompound tileNbt;
    @SuppressWarnings("WeakerAccess")
    protected Rotation tileRotation = Rotation.NONE;
    @SuppressWarnings("WeakerAccess")
    protected Block placeBlock;
    @SuppressWarnings("WeakerAccess")
    protected final Set<BlockPos> updateBlockOffsets = new HashSet<>();
    @SuppressWarnings("WeakerAccess")
    protected final Set<Block> canBeReplacedWithBlocks = new HashSet<>();

    @SuppressWarnings("unused")
    public static boolean predicate(SchematicBlockContext context) {
        if (context.blockState.getBlock().isAir(context.blockState, null, null)) {
            return false;
        }
        ResourceLocation registryName = context.block.getRegistryName();
        // noinspection ConstantConditions
        return registryName != null &&
            RulesLoader.READ_DOMAINS.contains(registryName.func_110624_b()) &&
            RulesLoader.getRules(
                context.blockState,
                context.block.hasTileEntity(context.blockState) && context.world.getTileEntity(context.pos) != null
                    ? context.world.getTileEntity(context.pos).serializeNBT()
                    : null
            ).stream()
                .noneMatch(rule -> rule.ignore);
    }

    @SuppressWarnings({"unused", "WeakerAccess"})
    protected void setRequiredBlockOffsets(SchematicBlockContext context, Set<JsonRule> rules) {
        requiredBlockOffsets.clear();
        rules.stream()
            .map(rule -> rule.requiredBlockOffsets)
            .filter(Objects::nonNull)
            .flatMap(Collection::stream)
            .forEach(requiredBlockOffsets::add);
        if (context.block instanceof BlockFalling) {
            requiredBlockOffsets.add(new BlockPos(0, -1, 0));
        }
    }

    @SuppressWarnings({"unused", "WeakerAccess"})
    protected void setBlockState(SchematicBlockContext context, Set<JsonRule> rules) {
        blockState = context.blockState;
    }

    @SuppressWarnings({"unused", "WeakerAccess"})
    protected void setIgnoredProperties(SchematicBlockContext context, Set<JsonRule> rules) {
        ignoredProperties.clear();
        rules.stream()
            .map(rule -> rule.ignoredProperties)
            .filter(Objects::nonNull)
            .flatMap(List::stream)
            .flatMap(propertyName ->
                context.blockState.getProperties().keySet().stream()
                    .filter(property -> property.getName().equals(propertyName))
            )
            .forEach(ignoredProperties::add);
    }

    @SuppressWarnings({"unused", "WeakerAccess"})
    protected void setTileNbt(SchematicBlockContext context, Set<JsonRule> rules) {
        tileNbt = null;
        if (context.block.hasTileEntity(context.blockState)) {
            TileEntity tileEntity = context.world.getTileEntity(context.pos);
            if (tileEntity != null) {
                tileNbt = tileEntity.serializeNBT();
            }
        }
    }

    @SuppressWarnings({"unused", "WeakerAccess"})
    protected void setPlaceBlock(SchematicBlockContext context, Set<JsonRule> rules) {
        placeBlock = rules.stream()
            .map(rule -> rule.placeBlock)
            .filter(Objects::nonNull)
            .findFirst()
            .map(Block::getBlockFromName)
            .orElse(context.block);
    }

    @SuppressWarnings({"unused", "WeakerAccess"})
    protected void setUpdateBlockOffsets(SchematicBlockContext context, Set<JsonRule> rules) {
        updateBlockOffsets.clear();
        if (rules.stream().map(rule -> rule.updateBlockOffsets).anyMatch(Objects::nonNull)) {
            rules.stream()
                .map(rule -> rule.updateBlockOffsets)
                .filter(Objects::nonNull)
                .flatMap(Collection::stream)
                .forEach(updateBlockOffsets::add);
        } else {
            Stream.of(EnumFacing.field_82609_l)
                .map(EnumFacing::func_176730_m)
                .map(BlockPos::new)
                .forEach(updateBlockOffsets::add);
            updateBlockOffsets.add(BlockPos.field_177992_a);
        }
    }

    @SuppressWarnings({"unused", "WeakerAccess"})
    protected void setCanBeReplacedWithBlocks(SchematicBlockContext context, Set<JsonRule> rules) {
        canBeReplacedWithBlocks.clear();
        rules.stream()
            .map(rule -> rule.canBeReplacedWithBlocks)
            .filter(Objects::nonNull)
            .flatMap(Collection::stream)
            .map(Block::func_149684_b)
            .forEach(canBeReplacedWithBlocks::add);
        canBeReplacedWithBlocks.add(context.block);
        canBeReplacedWithBlocks.add(placeBlock);
    }

    @Override
    public void init(SchematicBlockContext context) {
        // noinspection ConstantConditions
        Set<JsonRule> rules = RulesLoader.getRules(
            context.blockState,
            context.block.hasTileEntity(context.blockState) && context.world.getTileEntity(context.pos) != null
                ? context.world.getTileEntity(context.pos).serializeNBT()
                : null
        );
        setRequiredBlockOffsets /*   */(context, rules);
        setBlockState /*             */(context, rules);
        setIgnoredProperties /*      */(context, rules);
        setTileNbt /*                */(context, rules);
        setPlaceBlock /*             */(context, rules);
        setUpdateBlockOffsets /*     */(context, rules);
        setCanBeReplacedWithBlocks /**/(context, rules);
    }

    @Nonnull
    @Override
    public Set<BlockPos> getRequiredBlockOffsets() {
        return requiredBlockOffsets;
    }

    @Nonnull
    @Override
    public List<ItemStack> computeRequiredItems() {
        Set<JsonRule> rules = RulesLoader.getRules(blockState, tileNbt);
        List<List<RequiredExtractor>> collect = rules.stream()
            .map(rule -> rule.requiredExtractors)
            .filter(Objects::nonNull)
            .collect(Collectors.toList());
        return (
            collect.isEmpty()
                ? Stream.of(new RequiredExtractorItemFromBlock())
                : collect.stream().flatMap(Collection::stream)
        )
            .flatMap(requiredExtractor -> requiredExtractor.extractItemsFromBlock(blockState, tileNbt).stream())
            .filter(((Predicate<ItemStack>) ItemStack::func_190926_b).negate())
            .collect(Collectors.toList());
    }

    @Nonnull
    @Override
    public List<FluidStack> computeRequiredFluids() {
        Set<JsonRule> rules = RulesLoader.getRules(blockState, tileNbt);
        return rules.stream()
            .map(rule -> rule.requiredExtractors)
            .filter(Objects::nonNull)
            .flatMap(Collection::stream)
            .flatMap(requiredExtractor -> requiredExtractor.extractFluidsFromBlock(blockState, tileNbt).stream())
            .filter(Objects::nonNull)
            .collect(Collectors.toList());
    }

    @Override
    public SchematicBlockDefault getRotated(Rotation rotation) {
        SchematicBlockDefault schematicBlock = SchematicBlockManager.createCleanCopy(this);
        requiredBlockOffsets.stream()
            .map(blockPos -> blockPos.func_190942_a(rotation))
            .forEach(schematicBlock.requiredBlockOffsets::add);
        schematicBlock.blockState = blockState.func_185907_a(rotation);
        schematicBlock.ignoredProperties.addAll(ignoredProperties);
        schematicBlock.tileNbt = tileNbt;
        schematicBlock.tileRotation = tileRotation.func_185830_a(rotation);
        schematicBlock.placeBlock = placeBlock;
        updateBlockOffsets.stream()
            .map(blockPos -> blockPos.func_190942_a(rotation))
            .forEach(schematicBlock.updateBlockOffsets::add);
        schematicBlock.canBeReplacedWithBlocks.addAll(canBeReplacedWithBlocks);
        return schematicBlock;
    }

    @Override
    public boolean canBuild(World world, BlockPos blockPos) {
        return world.func_175623_d(blockPos);
    }

    @Override
    @SuppressWarnings("Duplicates")
    public boolean build(World world, BlockPos blockPos) {
        if (placeBlock == Blocks.field_150350_a) {
            return true;
        }
        world.field_72984_F.func_76320_a("prepare block");
        IBlockState newBlockState = blockState;
        if (placeBlock != blockState.func_177230_c()) {
            newBlockState = placeBlock.func_176223_P();
            for (IProperty<?> property : blockState.func_177227_a()) {
                if (newBlockState.func_177227_a().contains(property)) {
                    newBlockState = BlockUtil.copyProperty(
                        property,
                        newBlockState,
                        blockState
                    );
                }
            }
        }
        for (IProperty<?> property : ignoredProperties) {
            newBlockState = BlockUtil.copyProperty(
                property,
                newBlockState,
                placeBlock.func_176223_P()
            );
        }
        world.field_72984_F.func_76319_b();
        world.field_72984_F.func_76320_a("place block");
        boolean b = world.func_180501_a(blockPos, newBlockState, 11);
        world.field_72984_F.func_76319_b();
        if (b) {
            world.field_72984_F.func_76320_a("notify");
            updateBlockOffsets.stream()
                .map(blockPos::func_177971_a)
                .forEach(updatePos -> world.func_175685_c(updatePos, placeBlock, false));
            world.field_72984_F.func_76319_b();
            if (tileNbt != null && blockState.func_177230_c().hasTileEntity(blockState)) {
                world.field_72984_F.func_76320_a("prepare tile");
                Set<JsonRule> rules = RulesLoader.getRules(blockState, tileNbt);
                NBTTagCompound replaceNbt = rules.stream()
                    .map(rule -> rule.replaceNbt)
                    .filter(Objects::nonNull)
                    .map(NBTBase.class::cast)
                    .reduce(NBTUtilBC::merge)
                    .map(NBTTagCompound.class::cast)
                    .orElse(null);
                NBTTagCompound newTileNbt = new NBTTagCompound();
                tileNbt.func_150296_c().stream()
                    .map(key -> Pair.of(key, tileNbt.func_74781_a(key)))
                    .forEach(kv -> newTileNbt.func_74782_a(kv.getKey(), kv.getValue()));
                newTileNbt.func_74768_a("x", blockPos.func_177958_n());
                newTileNbt.func_74768_a("y", blockPos.func_177956_o());
                newTileNbt.func_74768_a("z", blockPos.func_177952_p());
                world.field_72984_F.func_76319_b();
                world.field_72984_F.func_76320_a("place tile");
                TileEntity tileEntity = TileEntity.func_190200_a(
                    world,
                    replaceNbt != null
                        ? (NBTTagCompound) NBTUtilBC.merge(newTileNbt, replaceNbt)
                        : newTileNbt
                );
                if (tileEntity != null) {
                    tileEntity.func_145834_a(world);
                    world.func_175690_a(blockPos, tileEntity);
                    if (tileRotation != Rotation.NONE) {
                        tileEntity.func_189667_a(tileRotation);
                    }
                }
                world.field_72984_F.func_76319_b();
            }
            return true;
        }
        return false;
    }

    @Override
    @SuppressWarnings("Duplicates")
    public boolean buildWithoutChecks(World world, BlockPos blockPos) {
        if (world.func_180501_a(blockPos, blockState, 0)) {
            if (tileNbt != null && blockState.func_177230_c().hasTileEntity(blockState)) {
                NBTTagCompound newTileNbt = new NBTTagCompound();
                tileNbt.func_150296_c().stream()
                    .map(key -> Pair.of(key, tileNbt.func_74781_a(key)))
                    .forEach(kv -> newTileNbt.func_74782_a(kv.getKey(), kv.getValue()));
                newTileNbt.func_74768_a("x", blockPos.func_177958_n());
                newTileNbt.func_74768_a("y", blockPos.func_177956_o());
                newTileNbt.func_74768_a("z", blockPos.func_177952_p());
                TileEntity tileEntity = TileEntity.func_190200_a(world, newTileNbt);
                if (tileEntity != null) {
                    tileEntity.func_145834_a(world);
                    world.func_175690_a(blockPos, tileEntity);
                    if (tileRotation != Rotation.NONE) {
                        tileEntity.func_189667_a(tileRotation);
                    }
                }
                return true;
            }
        }
        return false;
    }

    @Override
    public boolean isBuilt(World world, BlockPos blockPos) {
        return blockState != null &&
            canBeReplacedWithBlocks.contains(world.func_180495_p(blockPos).func_177230_c()) &&
            BlockUtil.blockStatesWithoutBlockEqual(blockState, world.func_180495_p(blockPos), ignoredProperties);
    }

    @Override
    public NBTTagCompound serializeNBT() {
        NBTTagCompound nbt = new NBTTagCompound();
        nbt.func_74782_a(
            "requiredBlockOffsets",
            NBTUtilBC.writeCompoundList(
                requiredBlockOffsets.stream()
                    .map(NBTUtil::func_186859_a)
            )
        );
        nbt.func_74782_a("blockState", NBTUtil.func_190009_a(new NBTTagCompound(), blockState));
        nbt.func_74782_a(
            "ignoredProperties",
            NBTUtilBC.writeStringList(
                ignoredProperties.stream()
                    .map(IProperty::func_177701_a)
            )
        );
        if (tileNbt != null) {
            nbt.func_74782_a("tileNbt", tileNbt);
        }
        nbt.func_74782_a("tileRotation", NBTUtilBC.writeEnum(tileRotation));
        nbt.func_74778_a("placeBlock", Block.field_149771_c.func_177774_c(placeBlock).toString());
        nbt.func_74782_a(
            "updateBlockOffsets",
            NBTUtilBC.writeCompoundList(
                updateBlockOffsets.stream()
                    .map(NBTUtil::func_186859_a)
            )
        );
        nbt.func_74782_a(
            "canBeReplacedWithBlocks",
            NBTUtilBC.writeStringList(
                canBeReplacedWithBlocks.stream()
                    .map(Block.field_149771_c::func_177774_c)
                    .map(Object::toString)
            )
        );
        return nbt;
    }

    @Override
    public void deserializeNBT(NBTTagCompound nbt) throws InvalidInputDataException {
        NBTUtilBC.readCompoundList(nbt.func_74781_a("requiredBlockOffsets"))
            .map(NBTUtil::func_186861_c)
            .forEach(requiredBlockOffsets::add);
        blockState = NBTUtil.func_190008_d(nbt.func_74775_l("blockState"));
        NBTUtilBC.readStringList(nbt.func_74781_a("ignoredProperties"))
            .map(propertyName ->
                blockState.func_177227_a().stream()
                    .filter(property -> property.func_177701_a().equals(propertyName))
                    .findFirst()
                    .orElse(null)
            )
            .forEach(ignoredProperties::add);
        if (nbt.func_74764_b("tileNbt")) {
            tileNbt = nbt.func_74775_l("tileNbt");
        }
        tileRotation = NBTUtilBC.readEnum(nbt.func_74781_a("tileRotation"), Rotation.class);
        placeBlock = Block.field_149771_c.func_82594_a(new ResourceLocation(nbt.func_74779_i("placeBlock")));
        NBTUtilBC.readCompoundList(nbt.func_74781_a("updateBlockOffsets"))
            .map(NBTUtil::func_186861_c)
            .forEach(updateBlockOffsets::add);
        NBTUtilBC.readStringList(nbt.func_74781_a("canBeReplacedWithBlocks"))
            .map(ResourceLocation::new)
            .map(Block.field_149771_c::getObject)
            .forEach(canBeReplacedWithBlocks::add);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        SchematicBlockDefault that = (SchematicBlockDefault) o;

        return requiredBlockOffsets.equals(that.requiredBlockOffsets) &&
            blockState.equals(that.blockState) &&
            ignoredProperties.equals(that.ignoredProperties) &&
            (tileNbt != null ? tileNbt.equals(that.tileNbt) : that.tileNbt == null) &&
            tileRotation == that.tileRotation &&
            placeBlock.equals(that.placeBlock) &&
            updateBlockOffsets.equals(that.updateBlockOffsets) &&
            canBeReplacedWithBlocks.equals(that.canBeReplacedWithBlocks);
    }

    @Override
    public int hashCode() {
        int result = requiredBlockOffsets.hashCode();
        result = 31 * result + blockState.hashCode();
        result = 31 * result + ignoredProperties.hashCode();
        result = 31 * result + (tileNbt != null ? tileNbt.hashCode() : 0);
        result = 31 * result + tileRotation.hashCode();
        result = 31 * result + placeBlock.hashCode();
        result = 31 * result + updateBlockOffsets.hashCode();
        result = 31 * result + canBeReplacedWithBlocks.hashCode();
        return result;
    }
}
