/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.lib.misc;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.annotation.Nonnull;

import net.minecraft.block.Block;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.item.ItemStack;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.NonNullList;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3d;
import net.minecraft.world.World;

import net.minecraftforge.items.IItemHandler;
import net.minecraftforge.items.IItemHandlerModifiable;

import buildcraft.api.inventory.IItemTransactor;
import buildcraft.api.transport.IInjectable;

import buildcraft.lib.inventory.ItemTransactorHelper;

public class InventoryUtil {
    // Drops

    public static void dropAll(World world, Vec3d vec, IItemHandlerModifiable handler) {
        dropAll(world, vec.field_72450_a, vec.field_72448_b, vec.field_72449_c, handler);
    }

    public static void dropAll(World world, BlockPos pos, IItemHandlerModifiable handler) {
        dropAll(world, pos.func_177958_n(), pos.func_177956_o(), pos.func_177952_p(), handler);
    }

    public static void dropAll(World world, double x, double y, double z, IItemHandlerModifiable handler) {
        for (int i = 0; i < handler.getSlots(); i++) {
            drop(world, x, y, z, handler.extractItem(i, Integer.MAX_VALUE, false));
        }
    }

    public static void dropAll(World world, BlockPos pos, NonNullList<ItemStack> toDrop) {
        for (ItemStack stack : toDrop) {
            if (stack == null) {
                throw new NullPointerException("Null stack!");
            }
            drop(world, pos, stack);
        }
    }

    public static void drop(World world, BlockPos pos, @Nonnull ItemStack stack) {
        Block.func_180635_a(world, pos, stack);
    }

    public static void drop(World world, Vec3d vec, @Nonnull ItemStack stack) {
        drop(world, vec.field_72450_a, vec.field_72448_b, vec.field_72449_c, stack);
    }

    public static void drop(World world, double x, double y, double z, @Nonnull ItemStack stack) {
        if (stack.func_190926_b()) {
            return;
        }
        EntityItem entity = new EntityItem(world, x, y, z, stack);
        world.func_72838_d(entity);
    }

    // Sending items around

    /** @return The leftover stack */
    @Nonnull
    public static ItemStack addToRandomInventory(World world, BlockPos pos, @Nonnull ItemStack stack) {
        if (stack.func_190926_b()) {
            return StackUtil.EMPTY;
        }
        List<EnumFacing> toTry = new ArrayList<>(6);
        Collections.addAll(toTry, EnumFacing.field_82609_l);
        Collections.shuffle(toTry);
        for (EnumFacing face : toTry) {
            TileEntity tile = world.func_175625_s(pos.func_177972_a(face));
            IItemTransactor transactor = ItemTransactorHelper.getTransactor(tile, face.func_176734_d());
            stack = transactor.insert(stack, false, false);
            if (stack.func_190926_b()) {
                return StackUtil.EMPTY;
            }
        }
        return stack;
    }

    /** Look around the tile given in parameter in all 6 position, tries to add the items to a random injectable tile
     * around. Will make sure that the location from which the items are coming from (identified by the from parameter)
     * isn't used again so that entities doesn't go backwards. Returns true if successful, false otherwise. */
    @Nonnull
    public static ItemStack addToRandomInjectable(World world, BlockPos pos, EnumFacing ignore, @Nonnull ItemStack stack) {
        if (stack.func_190926_b()) {
            return StackUtil.EMPTY;
        }
        List<EnumFacing> toTry = new ArrayList<>(6);
        Collections.addAll(toTry, EnumFacing.field_82609_l);
        Collections.shuffle(toTry);
        for (EnumFacing face : toTry) {
            if (face == ignore) continue;
            TileEntity tile = world.func_175625_s(pos.func_177972_a(face));
            IInjectable injectable = ItemTransactorHelper.getInjectable(tile, face.func_176734_d());
            stack = injectable.injectItem(stack, true, face.func_176734_d(), null, 0);
            if (stack.func_190926_b()) {
                return StackUtil.EMPTY;
            }
        }
        return stack;
    }

    /** Attempts to add the given stack to the best acceptor, in this order: {@link IItemHandler} instances,
     * {@link IInjectable} instances, and finally dropping it down on the ground. */
    public static void addToBestAcceptor(World world, BlockPos pos, EnumFacing ignore, @Nonnull ItemStack stack) {
        stack = addToRandomInjectable(world, pos, ignore, stack);
        stack = addToRandomInventory(world, pos, stack);
        drop(world, pos, stack);
    }

    /** Adds every stack from src to dst. Doesn't add empty stacks. */
    public static void addAll(IItemHandler src, NonNullList<ItemStack> dst) {
        for (int i = 0; i < src.getSlots(); i++) {
            ItemStack stack = src.getStackInSlot(i);
            if (!stack.func_190926_b()) {
                dst.add(stack);
            }
        }
    }

    // NBT migration
}
