/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.lib.misc;

import java.util.BitSet;
import java.util.EnumSet;
import java.util.Optional;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import com.google.common.collect.Sets;

import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTBase;
import net.minecraft.nbt.NBTTagByte;
import net.minecraft.nbt.NBTTagByteArray;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagDouble;
import net.minecraft.nbt.NBTTagIntArray;
import net.minecraft.nbt.NBTTagList;
import net.minecraft.nbt.NBTTagString;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3d;

import net.minecraftforge.common.util.Constants;

import buildcraft.api.core.BCLog;

public final class NBTUtilBC {
    @SuppressWarnings("WeakerAccess")
    public static final NBTTagCompound NBT_NULL = new NBTTagCompound();

    public static <N extends NBTBase> Optional<N> toOptional(N value) {
        return value == NBTUtilBC.NBT_NULL ? Optional.empty() : Optional.of(value);
    }

    public static NBTBase merge(NBTBase destination, NBTBase source) {
        if (source == null) {
            return null;
        }
        if (destination == null) {
            return source;
        }
        if (destination.func_74732_a() == Constants.NBT.TAG_COMPOUND && source.func_74732_a() == Constants.NBT.TAG_COMPOUND) {
            NBTTagCompound result = new NBTTagCompound();
            for (String key : Sets.union(
                ((NBTTagCompound) destination).func_150296_c(),
                ((NBTTagCompound) source).func_150296_c()
            )) {
                if (!((NBTTagCompound) source).func_74764_b(key)) {
                    result.func_74782_a(key, ((NBTTagCompound) destination).func_74781_a(key));
                } else if (((NBTTagCompound) source).func_74781_a(key) != NBT_NULL) {
                    if (!((NBTTagCompound) destination).func_74764_b(key)) {
                        result.func_74782_a(key, ((NBTTagCompound) source).func_74781_a(key));
                    } else {
                        result.func_74782_a(
                            key,
                            merge(
                                ((NBTTagCompound) destination).func_74781_a(key),
                                ((NBTTagCompound) source).func_74781_a(key)
                            )
                        );
                    }
                }
            }
            return result;
        }
        return source;
    }

    public static NBTTagCompound getItemData(@Nonnull ItemStack stack) {
        if (stack.func_190926_b()) {
            return new NBTTagCompound();
        }
        NBTTagCompound nbt = stack.func_77978_p();
        if (nbt == null) {
            nbt = new NBTTagCompound();
            stack.func_77982_d(nbt);
        }
        return nbt;
    }

    public static NBTTagIntArray writeBlockPos(BlockPos pos) {
        if (pos == null) {
            throw new NullPointerException("Cannot return a null NBTTag -- pos was null!");
        }
        return new NBTTagIntArray(new int[] { pos.func_177958_n(), pos.func_177956_o(), pos.func_177952_p() });
    }

    @SuppressWarnings("unused")
    public static NBTTagCompound writeBlockPosAsCompound(BlockPos pos) {
        if (pos == null) {
            throw new NullPointerException("Cannot return a null NBTTag -- pos was null!");
        }
        NBTTagCompound nbt = new NBTTagCompound();
        nbt.func_74768_a("x", pos.func_177958_n());
        nbt.func_74768_a("y", pos.func_177956_o());
        nbt.func_74768_a("z", pos.func_177952_p());
        return nbt;
    }

    @Nullable
    public static BlockPos readBlockPos(NBTBase base) {
        if (base == null) {
            return null;
        }
        switch (base.func_74732_a()) {
            case Constants.NBT.TAG_INT_ARRAY: {
                int[] array = ((NBTTagIntArray) base).func_150302_c();
                if (array.length == 3){
                    return new BlockPos(array[0], array[1], array[2]);
                }
                return null;
            }
            case Constants.NBT.TAG_COMPOUND: {
                NBTTagCompound nbt = (NBTTagCompound) base;
                BlockPos pos = null;
                if (nbt.func_74764_b("i")) {
                    int i = nbt.func_74762_e("i");
                    int j = nbt.func_74762_e("j");
                    int k = nbt.func_74762_e("k");
                    pos = new BlockPos(i, j, k);
                } else if (nbt.func_74764_b("x")) {
                    int x = nbt.func_74762_e("x");
                    int y = nbt.func_74762_e("y");
                    int z = nbt.func_74762_e("z");
                    pos = new BlockPos(x, y, z);
                } else if (nbt.func_74764_b("pos")) {
                    return readBlockPos(nbt.func_74781_a("pos"));
                } else {
                    BCLog.logger.warn("Attempted to read a block positions from a compound tag without the correct sub-tags! (" + base + ")", new Throwable());
                }
                return pos;
            }
        }
        BCLog.logger.warn("Attempted to read a block position from an invalid tag! (" + base + ")", new Throwable());
        return null;
    }

    public static NBTTagList writeVec3d(Vec3d vec3) {
        NBTTagList list = new NBTTagList();
        list.func_74742_a(new NBTTagDouble(vec3.field_72450_a));
        list.func_74742_a(new NBTTagDouble(vec3.field_72448_b));
        list.func_74742_a(new NBTTagDouble(vec3.field_72449_c));
        return list;
    }

    @Nullable
    public static Vec3d readVec3d(NBTBase nbt) {
        if (nbt instanceof NBTTagList) {
            return readVec3d((NBTTagList) nbt);
        }
        return null;
    }

    public static Vec3d readVec3d(NBTTagList list) {
        return new Vec3d(list.func_150309_d(0), list.func_150309_d(1), list.func_150309_d(2));
    }

    private static final String NULL_ENUM_STRING = "_NULL";

    public static <E extends Enum<E>> NBTBase writeEnum(E value) {
        if (value == null) {
            return new NBTTagString(NULL_ENUM_STRING);
        }
        return new NBTTagString(value.name());
    }

    public static <E extends Enum<E>> E readEnum(NBTBase nbt, Class<E> clazz) {
        if (nbt instanceof NBTTagString) {
            String value = ((NBTTagString) nbt).func_150285_a_();
            if (NULL_ENUM_STRING.equals(value)) {
                return null;
            }
            try {
                return Enum.valueOf(clazz, value);
            } catch (Throwable t) {
                // In case we didn't find the constant
                BCLog.logger.warn("Tried and failed to read the value(" + value + ") from " + clazz.getSimpleName(), t);
                return null;
            }
        } else if (nbt instanceof NBTTagByte) {
            byte value = ((NBTTagByte) nbt).func_150290_f();
            if (value < 0 || value >= clazz.getEnumConstants().length) {
                return null;
            } else {
                return clazz.getEnumConstants()[value];
            }
        } else if (nbt == null) {
            return null;
        } else {
            BCLog.logger.warn(new IllegalArgumentException("Tried to read an enum value when it was not a string! This is probably not good!"));
            return null;
        }
    }

    public static NBTBase writeDoubleArray(double[] data) {
        NBTTagList list = new NBTTagList();
        for (double d : data) {
            list.func_74742_a(new NBTTagDouble(d));
        }
        return list;
    }

    public static double[] readDoubleArray(NBTBase tag, int intendedLength) {
        double[] arr = new double[intendedLength];
        if (tag instanceof NBTTagList) {
            NBTTagList list = (NBTTagList) tag;
            for (int i = 0; i < list.func_74745_c() && i < intendedLength; i++) {
                arr[i] = list.func_150309_d(i);
            }
        }
        return arr;
    }

    /** Writes an {@link EnumSet} to an {@link NBTBase}. The returned type will either be {@link NBTTagByte} or
     * {@link NBTTagByteArray}.
     * 
     * @param clazz The class that the {@link EnumSet} is of. This is required as we have no way of getting the class
     *            from the set. */
    public static <E extends Enum<E>> NBTBase writeEnumSet(EnumSet<E> set, Class<E> clazz) {
        E[] constants = clazz.getEnumConstants();
        if (constants == null) throw new IllegalArgumentException("Not an enum type " + clazz);
        BitSet bitset = new BitSet();
        for (E e : constants) {
            if (set.contains(e)) {
                bitset.set(e.ordinal());
            }
        }
        byte[] bytes = bitset.toByteArray();
        if (bytes.length == 1) {
            return new NBTTagByte(bytes[0]);
        } else {
            return new NBTTagByteArray(bytes);
        }
    }

    public static <E extends Enum<E>> EnumSet<E> readEnumSet(NBTBase tag, Class<E> clazz) {
        E[] constants = clazz.getEnumConstants();
        if (constants == null) throw new IllegalArgumentException("Not an enum type " + clazz);
        byte[] bytes;
        if (tag instanceof NBTTagByte) {
            bytes = new byte[] { ((NBTTagByte) tag).func_150290_f() };
        } else if (tag instanceof NBTTagByteArray) {
            bytes = ((NBTTagByteArray) tag).func_150292_c();
        } else {
            bytes = new byte[] {};
            BCLog.logger.warn("[lib.nbt] Tried to read an enum set from " + tag);
        }
        BitSet bitset = BitSet.valueOf(bytes);
        EnumSet<E> set = EnumSet.noneOf(clazz);
        for (E e : constants) {
            if (bitset.get(e.ordinal())) {
                set.add(e);
            }
        }
        return set;
    }

    public static NBTTagList writeCompoundList(Stream<NBTTagCompound> stream) {
        NBTTagList list = new NBTTagList();
        stream.forEach(list::func_74742_a);
        return list;
    }

    public static Stream<NBTTagCompound> readCompoundList(NBTBase list) {
        if (list == null) {
            return Stream.empty();
        }
        if (!(list instanceof NBTTagList)) {
            throw new IllegalArgumentException();
        }
        return IntStream.range(0, ((NBTTagList) list).func_74745_c()).mapToObj(((NBTTagList) list)::func_150305_b);
    }

    public static NBTTagList writeStringList(Stream<String> stream) {
        NBTTagList list = new NBTTagList();
        stream.map(NBTTagString::new).forEach(list::appendTag);
        return list;
    }

    public static Stream<String> readStringList(NBTBase list) {
        if (list == null) {
            return Stream.empty();
        }
        if (!(list instanceof NBTTagList)) {
            throw new IllegalArgumentException();
        }
        return IntStream.range(0, ((NBTTagList) list).func_74745_c()).mapToObj(((NBTTagList) list)::func_150307_f);
    }
}
