/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team This Source Code Form is subject to the terms of the Mozilla
 * Public License, v. 2.0. If a copy of the MPL was not distributed with this file, You can obtain one at
 * https://mozilla.org/MPL/2.0/
 */

package buildcraft.lib.misc;

import java.util.Random;
import java.util.Set;
import java.util.function.Consumer;

import javax.annotation.Nullable;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;

import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumFacing.Axis;
import net.minecraft.util.Rotation;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3d;
import net.minecraft.util.math.Vec3i;

public class PositionUtil {
    @Nullable
    public static EnumFacing getDirectFacingOffset(BlockPos from, BlockPos to) {
        BlockPos diff = to.func_177973_b(from);
        boolean x = diff.func_177958_n() != 0;
        boolean y = diff.func_177956_o() != 0;
        boolean z = diff.func_177952_p() != 0;
        if (x && y || x && z || y && z) return null;
        if (x) return diff.func_177958_n() > 0 ? EnumFacing.EAST : EnumFacing.WEST;
        if (y) return diff.func_177956_o() > 0 ? EnumFacing.UP : EnumFacing.DOWN;
        if (z) return diff.func_177952_p() > 0 ? EnumFacing.SOUTH : EnumFacing.NORTH;
        return null;
    }

    public static Set<BlockPos> getCorners(BlockPos min, BlockPos max) {
        if (min == null || max == null) return ImmutableSet.of();
        if (min.equals(max)) return ImmutableSet.of(min);
        ImmutableSet.Builder<BlockPos> set = ImmutableSet.builder();
        set.add(min);
        set.add(new BlockPos(max.func_177958_n(), min.func_177956_o(), min.func_177952_p()));
        set.add(new BlockPos(min.func_177958_n(), max.func_177956_o(), min.func_177952_p()));
        set.add(new BlockPos(max.func_177958_n(), max.func_177956_o(), min.func_177952_p()));
        set.add(new BlockPos(min.func_177958_n(), min.func_177956_o(), max.func_177952_p()));
        set.add(new BlockPos(max.func_177958_n(), min.func_177956_o(), max.func_177952_p()));
        set.add(new BlockPos(min.func_177958_n(), max.func_177956_o(), max.func_177952_p()));
        set.add(max);
        return set.build();
    }

    private static int getBoxAxisCount(BlockPos min, BlockPos max, BlockPos pos) {
        if (min == null || max == null || pos == null) {
            return 0;
        }
        int same = 0;
        if (min.func_177958_n() == pos.func_177958_n() || max.func_177958_n() == pos.func_177958_n()) {
            same++;
        }
        if (min.func_177956_o() == pos.func_177956_o() || max.func_177956_o() == pos.func_177956_o()) {
            same++;
        }
        if (min.func_177952_p() == pos.func_177952_p() || max.func_177952_p() == pos.func_177952_p()) {
            same++;
        }
        return same;
    }

    /** Checks to see if the given position is a corner for the box given by min and max
     * 
     * @param min The minimum co-ordinate of the box
     * @param max The maximum co-ordinate of the box
     * @param pos The position to test
     * @return True if this position was on a corner, false if not. */
    public static boolean isCorner(BlockPos min, BlockPos max, BlockPos pos) {
        return getBoxAxisCount(min, max, pos) == 3;
    }

    /** Checks to see if the given position is on one of the edges of the box given by min and max
     * 
     * @param min The minimum co-ordinate of the box
     * @param max The maximum co-ordinate of the box
     * @param pos The position to test
     * @return True if this position was on an edge, false if not. */
    public static boolean isOnEdge(BlockPos min, BlockPos max, BlockPos pos) {
        return getBoxAxisCount(min, max, pos) >= 2;
    }

    /** Checks to see if the given position is on one of the faces of the box given by min and max
     * 
     * @param min The minimum co-ordinate of the box
     * @param max The maximum co-ordinate of the box
     * @param pos The position to test
     * @return True if this position was on a face, false if not. */
    public static boolean isOnFace(BlockPos min, BlockPos max, BlockPos pos) {
        return getBoxAxisCount(min, max, pos) >= 1;
    }

    public static boolean isNextTo(BlockPos one, BlockPos two) {
        BlockPos diff = one.func_177973_b(two);
        boolean x = diff.func_177958_n() == 1 || diff.func_177958_n() == -1;
        boolean y = diff.func_177956_o() == 1 || diff.func_177956_o() == -1;
        if (x && y) return false;
        boolean z = diff.func_177952_p() == 1 || diff.func_177952_p() == -1;
        if (y && z) return false;
        return x != z;
    }

    /** Finds a rotation that {@link #rotateFacing(EnumFacing, Axis, Rotation)} will use on "from" to get "to", with a
     * given axis around. */
    public static Rotation getRotatedFacing(EnumFacing from, EnumFacing to, Axis axis) {
        if (from.func_176740_k() == axis || to.func_176740_k() == axis) {
            throw new IllegalArgumentException("Cannot rotate around " + axis + " with " + from + " and " + to);
        }
        if (from == to) {
            return Rotation.NONE;
        }
        if (from.func_176734_d() == to) {
            return Rotation.CLOCKWISE_180;
        }
        if (from.func_176732_a(axis) == to) {
            return Rotation.CLOCKWISE_90;
        } else {
            return Rotation.COUNTERCLOCKWISE_90;
        }
    }

    /** Rotates a given {@link EnumFacing} by the given rotation, in a given axis. This relies on the behaviour defined
     * in {@link EnumFacing#rotateAround(Axis)}. */
    public static EnumFacing rotateFacing(EnumFacing from, Axis axis, Rotation rotation) {
        if (rotation == Rotation.NONE || rotation == null) {
            return from;
        }
        if (from.func_176740_k() == axis) {
            return from;
        } else if (rotation == Rotation.CLOCKWISE_180) {
            return from.func_176734_d();
        }

        if (rotation == Rotation.COUNTERCLOCKWISE_90) {
            // 270 is the same as 180 + 90
            // Vanilla gives us 90 for free.
            from = from.func_176734_d();
        }
        return from.func_176732_a(axis);
    }

    /** Rotates a given vector by the given rotation, in a given axis. This relies on the behaviour of
     * {@link #rotateFacing(EnumFacing, Axis, Rotation)}. */
    public static Vec3d rotateVec(Vec3d from, Axis axis, Rotation rotation) {
        Vec3d rotated = new Vec3d(0, 0, 0);

        double numEast = from.field_72450_a;
        double numUp = from.field_72448_b;
        double numSouth = from.field_72449_c;

        EnumFacing newEast = PositionUtil.rotateFacing(EnumFacing.EAST, axis, rotation);
        EnumFacing newUp = PositionUtil.rotateFacing(EnumFacing.UP, axis, rotation);
        EnumFacing newSouth = PositionUtil.rotateFacing(EnumFacing.SOUTH, axis, rotation);

        rotated = VecUtil.replaceValue(rotated, newEast.func_176740_k(), numEast * newEast.func_176743_c().func_179524_a());
        rotated = VecUtil.replaceValue(rotated, newUp.func_176740_k(), numUp * newUp.func_176743_c().func_179524_a());
        rotated = VecUtil.replaceValue(rotated, newSouth.func_176740_k(), numSouth * newSouth.func_176743_c().func_179524_a());

        return rotated;
    }

    /** Rotates a given position by the given rotation, in a given axis. This relies on the behaviour of
     * {@link #rotateFacing(EnumFacing, Axis, Rotation)}. */
    public static BlockPos rotatePos(Vec3i from, Axis axis, Rotation rotation) {
        BlockPos rotated = new BlockPos(0, 0, 0);

        int numEast = from.func_177958_n();
        int numUp = from.func_177956_o();
        int numSouth = from.func_177952_p();

        EnumFacing newEast = PositionUtil.rotateFacing(EnumFacing.EAST, axis, rotation);
        EnumFacing newUp = PositionUtil.rotateFacing(EnumFacing.UP, axis, rotation);
        EnumFacing newSouth = PositionUtil.rotateFacing(EnumFacing.SOUTH, axis, rotation);

        rotated = VecUtil.replaceValue(rotated, newEast.func_176740_k(), numEast * newEast.func_176743_c().func_179524_a());
        rotated = VecUtil.replaceValue(rotated, newUp.func_176740_k(), numUp * newUp.func_176743_c().func_179524_a());
        rotated = VecUtil.replaceValue(rotated, newSouth.func_176740_k(), numSouth * newSouth.func_176743_c().func_179524_a());

        return rotated;
    }

    public static LineSkewResult findLineSkewPoint(Line line, Vec3d start, Vec3d direction) {
        double ia = 0, ib = 1;
        double da = 0, db = 0;
        double id = 0.5;
        Vec3d va, vb;

        Vec3d best = null;
        for (int i = 0; i < 10; i++) {
            Vec3d a = line.interpolate(ia);
            Vec3d b = line.interpolate(ib);
            va = closestPointOnLineToPoint(a, start, direction);
            vb = closestPointOnLineToPoint(b, start, direction);
            da = a.func_72436_e(va);
            db = b.func_72436_e(vb);
            if (da < db) {
                // We work out the square root at the end to get the actual distance
                best = a;
                ib -= id;
            } else /* if (db < da) */ {
                // We work out the square root at the end to get the actual distance
                best = b;
                ia += id;
            }
            id /= 2.0;
        }
        return new LineSkewResult(best, Math.sqrt(Math.min(da, db)));
    }

    public static class LineSkewResult {
        public final Vec3d closestPos;
        public final double distFromLine;

        public LineSkewResult(Vec3d closestPos, double distFromLine) {
            this.closestPos = closestPos;
            this.distFromLine = distFromLine;
        }
    }

    public static Vec3d closestPointOnLineToPoint(Vec3d point, Vec3d linePoint, Vec3d lineVector) {
        Vec3d v = lineVector.func_72432_b();
        Vec3d p1 = linePoint;
        Vec3d p2 = point;

        // Its maths. Its allowed to deviate from normal naming rules.
        Vec3d p2_minus_p1 = p2.func_178788_d(p1);
        double _dot_v = VecUtil.dot(p2_minus_p1, v);
        Vec3d _scale_v = VecUtil.scale(v, _dot_v);
        return p1.func_178787_e(_scale_v);
    }

    public static class Line {
        public final Vec3d start, end;

        public Line(Vec3d start, Vec3d end) {
            this.start = start;
            this.end = end;
        }

        public static Line createLongLine(Vec3d start, Vec3d direction) {
            return new Line(start, VecUtil.scale(direction, 1024));
        }

        public Vec3d interpolate(double interp) {
            return VecUtil.scale(start, 1 - interp).func_178787_e(VecUtil.scale(end, interp));
        }
    }

    /** Returns a list of all the block positions on the edge of the given box. */
    public static ImmutableList<BlockPos> getAllOnEdge(BlockPos min, BlockPos max) {
        ImmutableList.Builder<BlockPos> list = ImmutableList.builder();
        boolean addX = max.func_177958_n() != min.func_177958_n();
        boolean addY = max.func_177956_o() != min.func_177956_o();
        boolean addZ = max.func_177952_p() != min.func_177952_p();
        if (addX & addY & addZ) {
            return getAllOnEdgeFull(min, max);
        }
        for (int x = min.func_177958_n(); x <= max.func_177958_n(); x++) {
            list.add(new BlockPos(x, min.func_177956_o(), min.func_177952_p()));
            if (addY) {
                list.add(new BlockPos(x, max.func_177956_o(), min.func_177952_p()));
                if (addZ) {
                    list.add(new BlockPos(x, max.func_177956_o(), max.func_177952_p()));
                }
            } else if (addZ) {
                list.add(new BlockPos(x, min.func_177956_o(), max.func_177952_p()));
            }
        }
        if (addY) {
            for (int y = min.func_177956_o() + 1; y < max.func_177956_o(); y++) {
                list.add(new BlockPos(min.func_177958_n(), y, min.func_177952_p()));
                if (addX) {
                    list.add(new BlockPos(max.func_177958_n(), y, min.func_177952_p()));
                    if (addZ) {
                        list.add(new BlockPos(max.func_177958_n(), y, max.func_177952_p()));
                    }
                } else if (addZ) {
                    list.add(new BlockPos(min.func_177958_n(), y, max.func_177952_p()));
                }
            }
        }
        if (addZ) {
            for (int z = min.func_177952_p() + 1; z < max.func_177952_p(); z++) {
                list.add(new BlockPos(min.func_177958_n(), min.func_177956_o(), z));
                if (addX) {
                    list.add(new BlockPos(max.func_177958_n(), min.func_177956_o(), z));
                    if (addY) {
                        list.add(new BlockPos(max.func_177958_n(), max.func_177956_o(), z));
                    }
                } else {
                    list.add(new BlockPos(min.func_177958_n(), max.func_177956_o(), z));
                }
            }
        }
        return list.build();
    }

    private static ImmutableList<BlockPos> getAllOnEdgeFull(BlockPos min, BlockPos max) {
        ImmutableList.Builder<BlockPos> list = ImmutableList.builder();
        for (int x = min.func_177958_n(); x <= max.func_177958_n(); x++) {
            list.add(new BlockPos(x, min.func_177956_o(), min.func_177952_p()));
            list.add(new BlockPos(x, max.func_177956_o(), min.func_177952_p()));
            list.add(new BlockPos(x, max.func_177956_o(), max.func_177952_p()));
            list.add(new BlockPos(x, min.func_177956_o(), max.func_177952_p()));
        }
        for (int y = min.func_177956_o() + 1; y < max.func_177956_o(); y++) {
            list.add(new BlockPos(min.func_177958_n(), y, min.func_177952_p()));
            list.add(new BlockPos(max.func_177958_n(), y, min.func_177952_p()));
            list.add(new BlockPos(max.func_177958_n(), y, max.func_177952_p()));
            list.add(new BlockPos(min.func_177958_n(), y, max.func_177952_p()));
        }
        for (int z = min.func_177952_p() + 1; z < max.func_177952_p(); z++) {
            list.add(new BlockPos(min.func_177958_n(), min.func_177956_o(), z));
            list.add(new BlockPos(max.func_177958_n(), min.func_177956_o(), z));
            list.add(new BlockPos(max.func_177958_n(), max.func_177956_o(), z));
            list.add(new BlockPos(min.func_177958_n(), max.func_177956_o(), z));
        }
        return list.build();
    }

    /** Returns a list of all the block positions between from and to (mostly).
     * <p>
     * Does not return the "from" co-ordinate, but does include the "to" co-ordinate (provided that from does not equal
     * to) */
    public static ImmutableList<BlockPos> getAllOnPath(BlockPos from, BlockPos to) {
        ImmutableList.Builder<BlockPos> interp = ImmutableList.builder();

        forAllOnPath(from, to, interp::add);

        return interp.build();
    }

    public static void forAllOnPath(BlockPos from, BlockPos to, Consumer<BlockPos> iter) {

        final BlockPos difference = to.func_177973_b(from);

        final int ax = Math.abs(difference.func_177958_n());
        final int ay = Math.abs(difference.func_177956_o());
        final int az = Math.abs(difference.func_177952_p());

        int count = ax + ay + az;
        BlockPos current = from;
        final int ddx = difference.func_177958_n() > 0 ? 1 : -1;
        final int ddy = difference.func_177956_o() > 0 ? 1 : -1;
        final int ddz = difference.func_177952_p() > 0 ? 1 : -1;

        // start from 1/2 in a block
        // (as we want to compare to the centre of blocks rather than the lower corner)
        int dx = count / 2;
        int dy = count / 2;
        int dz = count / 2;

        for (int j = 0; j < count; j++) {
            dx += ax;
            dy += ay;
            dz += az;
            boolean changed = false;
            if (dx >= count) {
                changed = true;
                dx -= count;
                current = current.func_177982_a(ddx, 0, 0);
            }
            if (dy >= count) {
                changed = true;
                dy -= count;
                current = current.func_177982_a(0, ddy, 0);
            }
            if (dz >= count) {
                changed = true;
                dz -= count;
                current = current.func_177982_a(0, 0, ddz);
            }

            if (changed) {
                iter.accept(current);
            }
        }
    }

    public static void forAllOnPath2d(int a1, int b1, int a2, int b2, PathIterator2d iter) {
        // Find the smallest number 'm' and smallest number 'o'
        // such that a * m + o = b
        // then draw a straight line (1, m)

        // First swap a with b if b is smaller than a

        int diff_a = a2 - a1;
        int diff_b = b2 - b1;

        int max_a = Math.abs(diff_a);
        int max_b = Math.abs(diff_b);

        int size_a = max_a + 1;
        int size_b = max_b + 1;

        int mult_a = diff_a > 0 ? 1 : -1;
        int mult_b = diff_b > 0 ? 1 : -1;

        boolean reverse = false;
        int multiplier;
        int offset;

        if (size_a == size_b) {
            multiplier = 1;
            offset = 0;
        } else {
            if (size_a > size_b) {
                int temp = size_a;
                size_a = size_b;
                size_b = temp;
                reverse = true;
            }
            multiplier = size_b / size_a;
            offset = size_b % size_a;
        }

        // Offset is distributed from the start of the line
        // Which is wrong atm -- need to distribute it across the line

        int normalLength = multiplier;
        int currentOffsetA = 0;
        int currentOffsetB = 0;
        int count = size_a;
        for (int i = 0; i < count; i++) {
            int length = normalLength;
            if (i < offset) {
                length++;
            }
            for (int l = 0; l < length; l++) {
                if (reverse) {
                    iter.iterate(a1 + mult_a * currentOffsetB, b1 + mult_b * currentOffsetA);
                } else {
                    iter.iterate(a1 + mult_a * currentOffsetA, b1 + mult_b * currentOffsetB);
                }
                currentOffsetB++;
            }
            currentOffsetA++;
        }
    }

    public static void forAllOnArc2d(int a, int b, int degrees, PathIterator2d iter) {
        
    }

    @FunctionalInterface
    public interface PathIterator2d {
        void iterate(int a, int b);
    }

    public static BlockPos randomBlockPos(Random rand, BlockPos size) {
        return new BlockPos(//
            rand.nextInt(size.func_177958_n()), //
            rand.nextInt(size.func_177956_o()), //
            rand.nextInt(size.func_177952_p())//
        );
    }

    public static BlockPos randomBlockPos(Random rand, BlockPos min, BlockPos max) {
        return new BlockPos(//
            min.func_177958_n() + rand.nextInt(max.func_177958_n() - min.func_177958_n()), //
            min.func_177956_o() + rand.nextInt(max.func_177956_o() - min.func_177956_o()), //
            min.func_177952_p() + rand.nextInt(max.func_177952_p() - min.func_177952_p())//
        );
    }
}
